﻿// <copyright file="Subscriber.cs" company="ANUJ Technologies Inc.">
// Copyright (c) 2013 All Rights Reserved. You may use this source code as long as the copyright header is kept intact.
// </copyright>
// <author>Anuj Varma</author>
// <blog>http://www.anujvarma.com/category/technology/</blog>
// <date>11/19/2013</date>
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace PublishSubscribeCSharp
{
    // Publisher Class – EmployerPublisher – will publish the PayrollArrived event
    public class EmployerPublisher 
    {
        LinkedList<Employee> _listSubscribers = new LinkedList<Employee>();
        public LinkedList<Employee> ListSubscribers
        {
            get { return _listSubscribers; }
            set { _listSubscribers = value; }
        }

        // For events to be treated as 'fields', they need to be of type 'delegate'.
        // First, define the delegate type for the event 
        public delegate void PayrollArrivedHandler(object sender, PublisherEventArgs e);

        // public event declaration, a 'delegate' type for the event must be declared first
        // initialize to empty delegate to avoid Gotcha 4 - race condition      
        public event PayrollArrivedHandler PayrollArrived = delegate {} ;

        // add new subscriber
        public void AddSubscriber(Employee em)
        {
            // Add new subscriber to the TOP of the linked list - this way it will not interfere with the notification in progress
            this.ListSubscribers.AddFirst(em);

            // Subscribe the new subscriber to the event.
            // The coupling between the publisher publishing and the subscriber subscribing is via the common delegate
            this.PayrollArrived += em.OnPayrollArrival;
        }

        // This event firing is synchronous - i.e. the publisher instance is tied down until the subscriber finishes processing OnPayrollArrival
        public void FireEvent(DateTime dayToFireEvent)
        {
            PublisherEventArgs args = new PublisherEventArgs();
            args.DateToFireEvent = dayToFireEvent;

            //  possible race event, payrollarrived can become null if another subscriber on another thread unsubscribes from this event
            if (PayrollArrived != null) // WILL THROW NULL REFERENCE IN THE CASE OF RACE CONDITION
                PayrollArrived(this, args);
        }

        // This is an asynchronous version of the same event-firing - just fire it - 
        // and do not CARE about how long the subscriber takes - since subscriber processing is on a different thread.
        // This works around the 'Badly behaved subscriber - Gotcha 1'.
        public void FireEventAsynchronous(DateTime dayToFireEvent)
        {
            if (PayrollArrived != null)
            {
                PublisherEventArgs args = new PublisherEventArgs();
                // We have more than one subscriber - employee 1 and empoyee 2. Each one subscribes via its own delegate. 
                // so - we need to loop over all the delegates

                Delegate[] delegates = PayrollArrived.GetInvocationList();
                foreach (Delegate del in delegates)
                {
                    PayrollArrivedHandler handler = (PayrollArrivedHandler) del;
                    handler.BeginInvoke(this, args, null, null);
                }
            }
        }

        public void RemoveSubscriber(Employee em)
        {
            this.PayrollArrived -= em.OnPayrollArrival;
        }
    }

    /// <summary>
    /// This makes it easy to pass values between publisher and subscriber, since this is common to 
    /// both (both use the same delegate PayrollArrivedHandler(object sender, PublisherEventArgs e);
    /// </summary>
    public class PublisherEventArgs : EventArgs
    {
        private DateTime _dateToFireEvent;

        public DateTime DateToFireEvent
        {
            get { return _dateToFireEvent; }
            set { _dateToFireEvent = value; }
        }
    }
}
